/*
* avc.c - AV/C Tape Recorder target implementation
*
* Copyright Dan Dennedy <dan@dennedy.org>
* Modified to confirm with JVC DVHS profile
*                               by Tim Coppefield <timecop@japan.co.jp>
* 
* Inspired by virtual_vcr from Bonin Franck <boninf@free.fr>
* 
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software Foundation,
* Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <sys/types.h>
#include <sys/time.h>

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>

#include <libraw1394/raw1394.h>
#include <libraw1394/csr.h>
#include <libavc1394/avc1394.h>

#include "avc.h"
#include "support.h"

unsigned char g_power = AVC1394_VCR_POWER_ON;

unsigned char g_transport_mode = AVC1394_VCR_CMD_WIND;
unsigned char g_transport_state = AVC1394_VCR_OPERAND_WIND_STOP;
// byte_t g_connect_block[5] = { 0xff, 0x00, 0x20, 0x00, 0x00 };
byte_t g_connect_block[5] = { 0x00, 0x00, 0x00, 0x00, 0x00 };

/* static functions */
static int avc_unit_control(avc1394_cmd_rsp *req);
static int avc_unit_status(avc1394_cmd_rsp *req);
static int avc_unit_inquiry(avc1394_cmd_rsp *req);

static int avc_subunit_control(avc1394_cmd_rsp *req);
static int avc_subunit_status(avc1394_cmd_rsp *req);
static int avc_subunit_inquiry(avc1394_cmd_rsp *req);

/*
 * Primary AV/C target callback function
 * Supports main UNIT + TAPE RECORDER SUBUNIT
 *
 */
int avc_command_handler(avc1394_cmd_rsp *req)
{
    char *command_type[] = { "CONTROL", "STATUS ", "SPECINQ", "NOTIFY ", "GENINQ " };

    eprintf("%s -> %s : %s : %02x\n",
            command_type[req->status],
	    (req->subunit_type == 0x1F) ? "UNIT   " : "SUBUNIT",
	    avc_command_name(req->opcode),
            req->operand[0]);

    /* process commands directed towards the main unit */
    if (req->subunit_type == AVC1394_SUBUNIT_UNIT) {
	/* UNIT */
	switch (req->status) {
	    case AVC1394_CTYP_CONTROL:
		return avc_unit_control(req);
		break;

	    case AVC1394_CTYP_STATUS:
		return avc_unit_status(req);
		break;

	    case AVC1394_CTYP_GENERAL_INQUIRY:
		return avc_unit_inquiry(req);
		break;
	}
    } else if (req->subunit_type == AVC1394_SUBUNIT_TAPE_RECORDER) {
	/* SUBINIT */
	switch (req->status) {
	    case AVC1394_CTYP_CONTROL:
		return avc_subunit_control(req);
		break;

	    case AVC1394_CTYP_STATUS:
		return avc_subunit_status(req);
		break;

	    case AVC1394_CTYP_GENERAL_INQUIRY:
		return avc_subunit_inquiry(req);
		break;
	}
    }

    return 0;
}


/*
 * AV/C Unit Control handler
 *
 * Mandatory commands:
 * OK: Implemented NG: Not Implemented HH: Hack
 *
 * CONNECT
 * INPUT_PLUG_SIGNAL_FORMAT
 * OUTPUT_PLUG_SIGNAL_FORMAT
 * OK	POWER
 */
static int avc_unit_control(avc1394_cmd_rsp *req)
{
    switch (req->opcode) {
	case AVC1394_CMD_CONNECT:
	    req->opcode = AVC1394_RESP_ACCEPTED;
	    g_connect_block[0] = req->operand[1]; // source subunit
            g_connect_block[1] = req->operand[2]; // source plug
            g_connect_block[2] = req->operand[3]; // dest subunit
            g_connect_block[3] = req->operand[4]; // dest plug
#if 0
	    req->opcode = AVC1394_RESP_ACCEPTED;
	    printf("AVC1394_CMD_CONNECT: %d (%02x:%02x) -> %d (%02x:%02x)\n",
		req->operand[2],
		req->operand[1] >> 3, req->operand[1] & 0x7,
		req->operand[4], req->operand[3] >> 3, req->operand[3] & 0x7);
#endif
	    break;

	case AVC1394_CMD_INPUT_PLUG_SIGNAL_FORMAT:


	    break;

	case AVC1394_CMD_OUTPUT_PLUG_SIGNAL_FORMAT:


	    break;

	case AVC1394_CMD_POWER:
	    /* make sure input is valid */
	    if (req->operand[0] == AVC1394_VCR_POWER_ON || req->operand[0] == AVC1394_VCR_POWER_OFF) {
                req->opcode = AVC1394_RESP_ACCEPTED;
		g_power = req->operand[0];
	    } else {
		req->opcode = AVC1394_RESP_REJECTED;
	    }
	    break;

	default:
	    eprintf("Unit Control Command 0x%02x not supported\n", req->opcode);
	    return 0;
	    break;
    }

    return 1;
}

/*
 * AV/C Unit Status handler
 *
 * Mandatory commands:
 * OK: Implemented NG: Not Implemented HH: Hack
 *
 * CONNECT
 * INPUT_PLUG_SIGNAL_FORMAT
 * OUTPUT_PLUG_SIGNAL_FORMAT
 * OK	SUBUNIT_INFO
 * OK	UNIT_INFO
 * OK	POWER
 */
static int avc_unit_status(avc1394_cmd_rsp *req)
{
    int page = 0;
    int rv = 1;

    switch (req->opcode) {
	case AVC1394_CMD_CONNECT:
	    req->status = AVC1394_RESP_STABLE;
	    if (req->operand[4] == 0xFE) {
		eprintf("                     Requesting info for SOURCE plug\n");
            } else {
                eprintf("                     Requesting info for DESTINATION plug\n");
                req->operand[0] = 0x3F << 2;
                req->operand[1] = g_connect_block[0];
                req->operand[2] = g_connect_block[1];
                req->operand[3] = g_connect_block[2];
                req->operand[4] = g_connect_block[3];
            }
	    break;

	case AVC1394_CMD_INPUT_PLUG_SIGNAL_FORMAT:

	    break;

	case AVC1394_CMD_OUTPUT_PLUG_SIGNAL_FORMAT:

	    break;


	case AVC1394_CMD_SUBUNIT_INFO:
	    page = (req->operand[0] >> 4) & 0x7;
	    if (page == 0) {
		req->status = AVC1394_RESP_STABLE;
		req->operand[0] = page << 4 | AVC1394_OPERAND_UNIT_INFO_EXTENSION_CODE;
		req->operand[1] = AVC1394_SUBUNIT_TAPE_RECORDER << 3;
		req->operand[2] = 0xFF; /* terminator */
	    } else {
		rv = 0;
	    }
	    break;

	case AVC1394_CMD_UNIT_INFO:
	    req->status = AVC1394_RESP_STABLE;
	    req->operand[0] = AVC1394_OPERAND_UNIT_INFO_EXTENSION_CODE;
	    req->operand[1] = AVC1394_SUBUNIT_TAPE_RECORDER << 3;
	    /* company_ID */
	    req->operand[2] = 0;
	    req->operand[3] = 0;
	    req->operand[4] = 0;
	    break;

	case AVC1394_CMD_POWER:
	    req->status = AVC1394_RESP_STABLE;
	    req->operand[0] = g_power;
	    break;

	default:
	    rv = 0;
	    break;
    }

    return rv;
}

/*
 * AV/C Unit Notify handler
 *
 * Mandatory commands:
 * OK: Implemented NG: Not Implemented HH: Hack
 *
 * OK	POWER
 */
static int avc_unit_inquiry(avc1394_cmd_rsp *req)
{
    req->status = AVC1394_RESP_NOT_IMPLEMENTED;

    switch (req->opcode) {
    case AVC1394_CMD_POWER:
	req->status = AVC1394_RESP_IMPLEMENTED;
	break;
    }
    return 1;
}

/*
 * AV/C SubUnit Control handler
 *
 * Mandatory commands:
 * OK: Implemented NG: Not Implemented HH: Hack
 *
 * OK	INPUT_SIGNAL_MODE
 * OK	OUTPUT_SIGNAL_MODE
 * NG	PLAY
 * HH	RECORD
 * RECORDING_SPEED
 * RELATIVE_TIME_COUNTER
 * TAPE_RECORDING_FORMAT
 * WIND
 */
static int avc_subunit_control(avc1394_cmd_rsp *req)
{
    int rv = 1;

    req->status = AVC1394_RESP_REJECTED;

    switch (req->opcode) {
	case AVC1394_VCR_CMD_INPUT_SIGNAL_MODE:
	    req->status = AVC1394_RESP_ACCEPTED;
	    eprintf("Setting INPUT signal mode to 0x%02x\n", req->operand[0]);
	    break;

	case AVC1394_VCR_CMD_OUTPUT_SIGNAL_MODE:
	    req->status = AVC1394_RESP_ACCEPTED;
	    eprintf("Setting OUTPUT signal mode to 0x%02x\n", req->operand[0]);
	    break;

	case AVC1394_VCR_CMD_RECORD:
	    req->status = AVC1394_RESP_ACCEPTED;
	    /* rv = avc_record(req); */
	    break;

	case AVC1394_VCR_CMD_RECORDING_SPEED:

	    break;

	case AVC1394_VCR_CMD_RELATIVE_TIME_COUNTER:

	    break;

	case AVC1394_VCR_CMD_TAPE_RECORDING_FORMAT:

	    break;

	case AVC1394_VCR_CMD_WIND:

	    break;

    }

    return rv;

#if 0
    switch (cr->opcode) {
    case AVC1394_VCR_CMD_RECORD:
	switch (cr->operand[0]) {
	case AVC1394_VCR_OPERAND_RECORD_RECORD:
	    g_transport_mode =
		AVC1394_GET_OPCODE
		(AVC1394_VCR_RESPONSE_TRANSPORT_STATE_RECORD);
	    g_transport_state = cr->operand[0];
	    cr->status = AVC1394_RESP_ACCEPTED;
	    printf("RECORD\n");
	    break;

	case AVC1394_VCR_OPERAND_RECORD_PAUSE:
	    g_transport_mode =
		AVC1394_GET_OPCODE
		(AVC1394_VCR_RESPONSE_TRANSPORT_STATE_RECORD);
	    g_transport_state = cr->operand[0];
	    cr->status = AVC1394_RESP_ACCEPTED;
	    printf("PAUSE RECORD\n");
	    break;

	default:
	    fprintf(stderr, "record mode 0x%02x non supported\n",
		    cr->operand[0]);
	    return 0;
	}
	break;
    case AVC1394_VCR_CMD_WIND:
	switch (cr->operand[0]) {
	case AVC1394_VCR_OPERAND_WIND_STOP:
	    g_transport_mode =
		AVC1394_GET_OPCODE
		(AVC1394_VCR_RESPONSE_TRANSPORT_STATE_WIND);
	    g_transport_state = cr->operand[0];
	    cr->status = AVC1394_RESP_ACCEPTED;
	    printf("STOP\n");
	    break;

	case AVC1394_VCR_OPERAND_WIND_REWIND:
	case AVC1394_VCR_OPERAND_WIND_HIGH_SPEED_REWIND:
	    g_transport_mode =
		AVC1394_GET_OPCODE(AVC1394_VCR_RESPONSE_TRANSPORT_STATE_WIND);
	    g_transport_state = AVC1394_VCR_OPERAND_WIND_REWIND;
	    cr->status = AVC1394_RESP_ACCEPTED;
	    printf("REWIND\n");
	    break;

	case AVC1394_VCR_OPERAND_WIND_FAST_FORWARD:
	    g_transport_mode =
		AVC1394_GET_OPCODE
		(AVC1394_VCR_RESPONSE_TRANSPORT_STATE_WIND);
	    g_transport_state = cr->operand[0];
	    cr->status = AVC1394_RESP_ACCEPTED;
	    printf("FAST FORWARD\n");
	    break;

	default:
	    fprintf(stderr, "wind mode 0x%02x non supported\n",
		    cr->operand[0]);
	    return 0;
	}
	break;

    case AVC1394_VCR_CMD_TAPE_RECORDING_FORMAT:
	cr->status = AVC1394_RESP_ACCEPTED;
	cr->operand[0] = 1 << 3;
	cr->operand[7] = 0xff;
	break;

    default:
	fprintf(stderr, "subunit control command 0x%02x non supported\n",
		cr->opcode);
	return 0;
    }
    return 1;
#endif
}

/*
 * AV/C SubUnit Status handler
 *
 * Mandatory commands:
 * OK: Implemented NG: Not Implemented HH: Hack
 *
 * OK	INPUT_SIGNAL_MODE	??
 * OK	MEDIUM_INFO
 * OK	OUTPUT_SIGNAL_MODE	??
 * OK	RECORDING_SPEED
 * HH	RELATIVE_TIME_COUNTER
 * OK	TAPE_PLAYBACK_FORMAT
 * OK	TAPE_RECORDING_FORMAT
 * OK	TRANSPORT_STATE
 */
static int avc_subunit_status(avc1394_cmd_rsp *req)
{
    int rv = 1;

    switch (req->opcode) {

    case AVC1394_VCR_CMD_INPUT_SIGNAL_MODE:
	req->status = AVC1394_RESP_STABLE;
	req->operand[0] = AVC1394_VCR_NORMAL_SIGNAL_MODE; // AVC1394_VCR_DVHS_SIGNAL_MODE;
	break;

    case AVC1394_VCR_CMD_OUTPUT_SIGNAL_MODE:
	req->status = AVC1394_RESP_STABLE;
	req->operand[0] = AVC1394_VCR_NORMAL_SIGNAL_MODE; // AVC1394_VCR_DVHS_SIGNAL_MODE;
	break;

    case AVC1394_VCR_CMD_MEDIUM_INFO:
	if (req->operand[0] == 0x7F && req->operand[1] == 0x7F) {
	    req->status = AVC1394_RESP_STABLE;
	    req->operand[0] = AVC1394_VCR_OPERAND_MEDIUM_INFO_VHS;
	    req->operand[1] = AVC1394_VCR_OPERAND_MEDIUM_INFO_DVHS;
	    break;
	} else {
	    req->status = AVC1394_RESP_REJECTED;
	    break;
	}

    case AVC1394_VCR_CMD_RECORDING_SPEED:
	req->status = AVC1394_RESP_STABLE;
	req->operand[0] = AVC1394_VCR_OPERAND_STANDARD_SPEED;
	break;

    case AVC1394_VCR_CMD_RELATIVE_TIME_COUNTER:
	req->status = AVC1394_RESP_STABLE;
	req->operand[0] = AVC1394_VCR_OPERAND_RECORDING_TIME_STATUS;
	req->operand[1] = 0x7F;
	req->operand[2] = 1;
	req->operand[3] = 0;
	req->operand[4] = 0;
	break;
    
    case AVC1394_VCR_CMD_TAPE_PLAYBACK_FORMAT:
	req->status = AVC1394_RESP_STABLE;
        req->operand[0] = 1 << 3;
        req->operand[1] = 0;
	// req->operand[1] = 1 << 5;
        req->operand[2] = 0x0;
        req->operand[3] = 0x0;
	req->operand[4] = 0;
        // req->operand[4] = 0x1;
        req->operand[5] = 0x0;
        req->operand[6] = 0x0;
        req->operand[7] = 0x0;
        req->operand[8] = 0x0;
	break;

    case AVC1394_VCR_CMD_TAPE_RECORDING_FORMAT:
	req->status = AVC1394_RESP_STABLE;
        req->operand[0] = 1 << 3;
        req->operand[1] = 1 << 5;
        req->operand[2] = 0x0;
        req->operand[3] = 0x0;
        req->operand[4] = 0x1;
        req->operand[5] = 0x0;
        req->operand[6] = 0x0;
        req->operand[7] = 0x0;
        req->operand[8] = 0x0;
	break;

    case AVC1394_VCR_CMD_TRANSPORT_STATE:
	req->status = AVC1394_RESP_STABLE;
	req->opcode = g_transport_mode;
	req->operand[0] = g_transport_state;
	break;

    default:
	eprintf("SubUnit Status Command 0x%02x not supported\n", req->opcode);
	rv = 0;
	break;
    }

    return rv;
}

/*
 * AV/C SubUnit Notify handler
 *
 * Mandatory commands:
 * OK: Implemented NG: Not Implemented HH: Hack
 *
 * TRANSPORT_STATE
 */
static int avc_subunit_inquiry(avc1394_cmd_rsp *req)
{
    int rv = 1;

    switch (req->opcode) {
	case AVC1394_VCR_CMD_TRANSPORT_STATE:
	    req->status = AVC1394_RESP_IMPLEMENTED;
	    break;
    }

    return rv;
}
